-----------------------------------------------------------------------------------------------
-- Client Lua Script for ChatAdvanced by Kertsz
-- Original ColorPicker by Jaslm and Tomber
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "Apollo"
require "ChatSystemLib"
require "GameLib"
 
-----------------------------------------------------------------------------------------------
-- ChatAdvanced Module Definition
-----------------------------------------------------------------------------------------------
local ChatAdvanced = {} 
local kcrInvalidColor = ApolloColor.new("InvalidChat")
local kcrValidColor = ApolloColor.new("white")

Key = {};

Key["backspace"] = 8;
Key["tab"] = 9;
Key["enter"] = 13;
Key["shift"] = 16;
Key["ctrl"] = 17;
Key["alt"] = 18;
Key["pause"] = 19;
Key["caps lock"] = 20
Key["esc"] = 27;
Key["spacebar"] = 32;
Key["page up"] = 33;
Key["page down"] = 34;
Key["end"] = 35;
Key["home"] = 36;
Key["left arrow"] = 37;
Key["up arrow"] = 38;
Key["right arrow"] = 39;
Key["down arrow"] = 40;
Key["insert"] = 45;
Key["delete"] = 46;
Key["0"] = 48;
Key["1"] = 49;
Key["2"] = 50;
Key["3"] = 51;
Key["4"] = 52;
Key["5"] = 53;
Key["6"] = 54;
Key["7"] = 55;
Key["8"] = 56;
Key["9"] = 57;
Key["a"] = 65;
Key["b"] = 66;
Key["c"] = 67;
Key["d"] = 68;
Key["e"] = 69;
Key["f"] = 70;
Key["g"] = 71;
Key["h"] = 72;
Key["i"] = 73;
Key["j"] = 74;
Key["k"] = 75;
Key["l"] = 76;
Key["m"] = 77;
Key["n"] = 78;
Key["o"] = 79;
Key["p"] = 80;
Key["q"] = 81;
Key["r"] = 82;
Key["s"] = 83;
Key["t"] = 84;
Key["u"] = 85;
Key["v"] = 86;
Key["w"] = 87;
Key["x"] = 88;
Key["y"] = 89;
Key["z"] = 90;
Key["lwindows"] = 91; -- Left Windows Key
Key["rwindows"] = 92; -- Right Windows Key
Key["appkey"] = 93; -- Application key - located between the right Windows and Ctrl keys on most keyboards
-- Numpad numbers start
Key["num 1"] = 97;
Key["num 2"] = 98;
Key["num 3"] = 99;
Key["num 4"] = 100;
Key["num 5"] = 101;
Key["num 6"] = 102;
Key["num 7"] = 103;
Key["num 8"] = 104;
Key["num 9"] = 105;
Key["nummult"] = 106;
Key["num plus"] = 107;
Key["num sub"] = 109;
Key["num decimal"] = 110;
Key["num divide"] = 111;
-- Numpad numbers end
Key["f1"] = 112;
Key["f2"] = 113;
Key["f3"] = 114;
Key["f4"] = 115;
Key["f5"] = 116;
Key["f6"] = 117;
Key["f7"] = 118;
Key["f8"] = 119;
Key["f9"] = 120;
Key["f10"] = 121;
Key["f11"] = 122;
Key["f12"] = 123;
Key["numlock"] = 144;
Key["scrolllock"] = 145;
Key[";"] = 186;
Key["="] = 187;
Key[","] = 188;
Key["-"] = 189;
Key["."] = 190;
Key["/"] = 191;
Key["`"] = 192;
Key["["] = 219;
Key["\\"] = 220;
Key["]" ] = 221;
Key["'"] = 221; 

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function ChatAdvanced:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    return o
end

function ChatAdvanced:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		--"Keybind"
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- ChatAdvanced OnLoad
-----------------------------------------------------------------------------------------------
function ChatAdvanced:OnLoad()
	self.strSaveVersion = "2.4"

	self.xmlDoc = XmlDoc.CreateFromFile("ChatAdvanced.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
		
	Apollo.RegisterSlashCommand("clearhistory", "ClearChatLogHistory", self)
	Apollo.RegisterSlashCommand("cls", "ClearChatLog", self)
	Apollo.RegisterEventHandler("SystemKeyDown", "OnSystemKeyDown", self)
	
	--Apollo.RegisterEventHandler("OpenSlashCommand", "OpenSlashCommand", self)
	
	-- Initialize ColorPickerAdvanced
	self.ColorPicker = ColorPickerAdvanced:new()
	self.ColorPicker:OnLoad()
	
	-- Load the defaults used in ChatLog 
	-- (the original is inaccessible in this scope, so it's duplicated here)	
	self.arChannelTypeToColor =
	{
		[ChatSystemLib.ChatChannel_Command] 		= { Channel = "ChannelCommand", 		},
		[ChatSystemLib.ChatChannel_System] 			= { Channel = "ChannelSystem", 			},
		[ChatSystemLib.ChatChannel_Debug] 			= { Channel = "ChannelDebug", 			},
		[ChatSystemLib.ChatChannel_Say] 			= { Channel = "ChannelSay", 			},
		[ChatSystemLib.ChatChannel_Yell] 			= { Channel = "ChannelShout", 			},
		[ChatSystemLib.ChatChannel_Whisper] 		= { Channel = "ChannelWhisper", 		},
		[ChatSystemLib.ChatChannel_Party] 			= { Channel = "ChannelParty", 			},
		[ChatSystemLib.ChatChannel_Emote] 			= { Channel = "ChannelEmote", 			},
		[ChatSystemLib.ChatChannel_AnimatedEmote] 	= { Channel = "ChannelEmote", 			},
		[ChatSystemLib.ChatChannel_Zone] 			= { Channel = "ChannelZone", 			},
		[ChatSystemLib.ChatChannel_ZonePvP] 		= { Channel = "ChannelPvP", 			},
		[ChatSystemLib.ChatChannel_Trade] 			= { Channel = "ChannelTrade",			},
		[ChatSystemLib.ChatChannel_Guild] 			= { Channel = "ChannelGuild", 			},
		[ChatSystemLib.ChatChannel_GuildOfficer] 	= { Channel = "ChannelGuildOfficer",	},
		[ChatSystemLib.ChatChannel_Society] 		= { Channel = "ChannelCircle2",			},
		[ChatSystemLib.ChatChannel_Custom] 			= { Channel = "ChannelCustom", 			},
		[ChatSystemLib.ChatChannel_NPCSay] 			= { Channel = "ChannelNPC", 			},
		[ChatSystemLib.ChatChannel_NPCYell] 		= { Channel = "ChannelNPC",		 		},
		[ChatSystemLib.ChatChannel_NPCWhisper]		= { Channel = "ChannelNPC", 			},
		[ChatSystemLib.ChatChannel_Datachron] 		= { Channel = "ChannelNPC", 			},
		[ChatSystemLib.ChatChannel_Combat] 			= { Channel = "ChannelGeneral", 		},
		[ChatSystemLib.ChatChannel_Realm] 			= { Channel = "ChannelSupport", 		},
		[ChatSystemLib.ChatChannel_Loot] 			= { Channel = "ChannelLoot", 			},
		[ChatSystemLib.ChatChannel_PlayerPath] 		= { Channel = "ChannelGeneral", 		},
		[ChatSystemLib.ChatChannel_Instance] 		= { Channel = "ChannelParty", 			},
		[ChatSystemLib.ChatChannel_WarParty] 		= { Channel = "ChannelWarParty",		},
		[ChatSystemLib.ChatChannel_WarPartyOfficer] = { Channel = "ChannelWarPartyOfficer", },
		[ChatSystemLib.ChatChannel_Advice] 			= { Channel = "ChannelAdvice", 			},
		[ChatSystemLib.ChatChannel_AccountWhisper] 	= { Channel = "ChannelAccountWisper", 	},
	}
		
	-- Connect to the stock ChatLog AddOn
	local ChatLog = ChatAdvanced:GetChatLog()
	ChatLog.ChatAdvanced = self
	
	ChatLog.crInvalidColor = ApolloColor.new("InvalidChat")
	ChatLog.crValidColor = ApolloColor.new("white")

	ChatLog.strColorChatRegular 	= "ff7fffb9"
	ChatLog.strColorChatShout		= "ffd9eef7"
	ChatLog.strColorChatRoleplay 	= "ff58e3b0"
	ChatLog.strBubbleFont 			= "CRB_Dialog"
	ChatLog.strDialogFont 			= "CRB_Dialog"
	ChatLog.strDialogFontRP 		= "CRB_Dialog_I"

	ChatLog.strGMIcon 		= "Icon_Windows_UI_GMIcon"
	ChatLog.arEvalColors =
	{
		[Item.CodeEnumItemQuality.Inferior] 		= "ItemQuality_Inferior",
		[Item.CodeEnumItemQuality.Average] 			= "ItemQuality_Average",
		[Item.CodeEnumItemQuality.Good] 			= "ItemQuality_Good",
		[Item.CodeEnumItemQuality.Excellent] 		= "ItemQuality_Excellent",
		[Item.CodeEnumItemQuality.Superb] 			= "ItemQuality_Superb",
		[Item.CodeEnumItemQuality.Legendary] 		= "ItemQuality_Legendary",
		[Item.CodeEnumItemQuality.Artifact]		 	= "ItemQuality_Artifact",
	}
	
	-- Post-Hook ChatLog:OnDocumentReady
	if(ChatLog.OnWindowManagementReady) then
		ChatLog.OriginalOnWindowManagementReady = ChatLog.OnWindowManagementReady
		function ChatLog.OnWindowManagementReady(tSelf)
			tSelf:OriginalOnWindowManagementReady()
			
			-- Adjust the ChatLog options panel	
			local wndOpts = tSelf.wndChatOptions:FindChild("ChatOptionsContent")
			wndOpts:FindChild("FontSizeSmall"):Show(false)
			wndOpts:FindChild("FontSizeMedium"):Show(false)
			wndOpts:FindChild("FontSizeLarge"):Show(false)
			local wndChatOptionsOverride = Apollo.LoadForm(tSelf.ChatAdvanced.xmlDoc, "ChatOptionsOverride", wndOpts, tSelf.ChatAdvanced)

			tSelf.ChatAdvanced:Ready()
		end
	else
		ChatLog.OriginalOnDocumentReady = ChatLog.OnDocumentReady
		function ChatLog.OnDocumentReady(tSelf)
			tSelf:OriginalOnDocumentReady()
			tSelf.ChatAdvanced:Ready()	
		end
	end
end

function ChatAdvanced:ClearChatLogHistory()
	self.arChatHistory = {}
end

function ChatAdvanced:ClearChatLog()
	self.arChatHistory = {}
	RequestReloadUI()
end


function ChatAdvanced:OnSystemKeyDown(iKey)
	if GameLib.GetPlayerUnit():IsInCombat() then
		return
	end

	for i=1,#self.arKeyBindings do
		if(iKey == Key[GameLib.GetInputKeyNameText(self.arKeyBindings[i]):lower()]) then
			self.timer = ApolloTimer.Create(0.1, false, "OnSlashCommand", self)
						
			return
		end
	end
end

function ChatAdvanced:OnSlashCommand()
	local wndEdit = self.ChatLog:HelperGetCurrentEditbox()
			
	wndEdit:SetFocus()
	wndEdit:SetText("/")
	wndEdit:SetSel(1, -1)
	wndEdit:GetParent():FindChild("InputType"):SetText("")

end

function ChatAdvanced:Ready()
	if not self.arChatColor then self:OnRestore() end
	for k,v in pairs(self.arChatColor) do
		self.ChatLog.arChatColor[k] = v
	end
	
	if(self.bRestored == false) then
		self.bRestored = true
		for _,tQueuedMessage in pairs(self.arChatHistory) do
			
			for key, wndChat in pairs(self.ChatLog.tChatWindows) do
				if wndChat:GetData().tViewedChannels[tQueuedMessage.eChannelType] then
					self.ChatLog.bQueuedMessages = true
					wndChat:GetData().tMessageQueue:Push(tQueuedMessage)
				end
			end
		end
	end

end

function ChatAdvanced:GetChatLog()
	if not self.ChatLog then 
		-- Check for some of the more common ChatLog replacements
		self.ChatLog = Apollo.GetAddon("ChatLog")
		if not self.ChatLog then
			self.ChatLog = Apollo.GetAddon("ChatFixed")
		end
		if not self.ChatLog then
			self.ChatLog = Apollo.GetAddon("BetterChatLog")
		end	
		if not self.ChatLog then
			self.ChatLog = Apollo.GetAddon("ImprovedChatLog")
		end
		if not self.ChatLog then
			self.ChatLog = Apollo.GetAddon("FixedChatLog")
		end
	end
	
	return self.ChatLog
end

-----------------------------------------------------------------------------------------------
-- ColorPicker Integration
-----------------------------------------------------------------------------------------------

-- When a channel's color box is clicked...
function ChatAdvanced:OnChannelColorClick(wndHandler, wndControl)
	tData = wndControl:GetData()
	local channel = tData.oChannel:GetName()..tData.oChannel:GetType()

	color = ChatAdvanced:GetChatLog().arChatColor[channel] or ChatAdvanced:GetChatLog().arChatColor[tData.oChannel:GetType()]
		
	self.ColorPicker.wndMain:FindChild("Title"):SetText(tData.oChannel:GetName().." Channel Options")
	self.ColorPicker.wndMain:FindChild("AllChannelsFont"):SetCheck(false)
	
	local strChannelName = self.arChannelNames[channel] or tData.oChannel:GetName()
	self.ColorPicker.wndMain:FindChild("ChannelAbbrev"):SetText(strChannelName)
	
	self.ColorPicker.wndMain:FindChild("SaveHistoryJoinLeave"):SetCheck(false)
	if self.arSaveChannel[channel] then
		if self.arSaveChannel[channel] > 0 then
			self.ColorPicker.wndMain:FindChild("SaveHistory"):SetCheck(true)
		end
		
		if self.arSaveChannel[channel] == 2 then
			self.ColorPicker.wndMain:FindChild("SaveHistoryJoinLeave"):SetCheck(true)
		end
	elseif self.arSaveChannel[tData.oChannel:GetType()] then
		if self.arSaveChannel[tData.oChannel:GetType()] > 0 then
			self.ColorPicker.wndMain:FindChild("SaveHistory"):SetCheck(true)
		end
		if self.arSaveChannel[tData.oChannel:GetType()] == 2 then
			self.ColorPicker.wndMain:FindChild("SaveHistoryJoinLeave"):SetCheck(true)
		end
	else
		self.ColorPicker.wndMain:FindChild("SaveHistory"):SetCheck(false)
	end

	if color then
		self.ColorPicker.AdjustCColor(CColor.new(color.r, color.g, color.b, color.a), false, ChatAdvanced.OnColorChanged, wndControl, self, { font = self.arFontOption[channel] })
	end
end

-- When the color changes in the ColorPicker
function ChatAdvanced:OnColorChanged(wndControl, newSettings, save)
	newColor = newSettings.color
	tData = wndControl:GetData()
	tPixieInfo = wndControl:GetPixieInfo(tData.nPixieId)
	tPixieInfo.cr = newColor
	wndControl:UpdatePixie(tData.nPixieId, tPixieInfo)
	
		
	if save then
		local eChannel = tData.oChannel:GetName()..tData.oChannel:GetType()

		self.ChatLog.arChatColor[eChannel] = newColor
		self.arChatColor[eChannel] = newColor
		self.arChannelNames[eChannel] = self.ColorPicker.wndMain:FindChild("ChannelAbbrev"):GetText()
		
		local strColor = "ff"..ChatAdvanced.Convert_RGB_To_String(newColor.r, newColor.g, newColor.b)
		self.arChannelTypeToColor[eChannel] = { Channel = strColor }
	    -- self.arChannelTypeToColor[tData.oChannel:GetType()] = { Channel = strColor }
		if (self.ColorPicker.wndMain:FindChild("SaveHistory"):IsChecked()) then
			self.arSaveChannel[eChannel] = 1
			if self.ColorPicker.wndMain:FindChild("SaveHistoryJoinLeave"):IsChecked() then
				self.arSaveChannel[eChannel] = 2
			end
		else
			self.arSaveChannel[eChannel] = 0
		end
		
		if self.ColorPicker.wndMain:FindChild("AllChannelsFont"):IsChecked() then
			for k,v in pairs(self.ChatLog.arChatColor) do
				self.arFontOption[k] = newSettings.font
			end
		else
			self.arFontOption[eChannel] = newSettings.font
		end
	end
end

function ChatAdvanced:SetFont(wndControl, font)
	
	--RequestReloadUI()
end

-- To help with scoping issues
function ChatAdvanced.Convert_RGB_To_String(r, g, b)
	return string.format("%02x%02x%02x", math.floor(r * 255 + 0.5), math.floor(g * 255 + 0.5), math.floor(b * 255 + 0.5))
end

-----------------------------------------------------------------------------------------------
-- ChatAdvanced OnSave / OnRestore
-----------------------------------------------------------------------------------------------
function ChatAdvanced:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return nil
	end
		
	-- Store ChatColors as hex strings
	local chatColors = {}
	for k,v in pairs(self.arChatColor) do
		chatColors[k] = "ff"..ChatAdvanced.Convert_RGB_To_String(v.r, v.g, v.b)
	end
	
	-- Only save the last 1000 entries
	history = {}
	size = #self.arChatHistory
	if size > 1000 then
		start = size - 1000
	else
		start = 1
	end
	
	for i=start, size, 1 do
		table.insert(history, self.arChatHistory[i])
	end

	local tSaved = {
		arChannelTypeToColor = self.arChannelTypeToColor,
		arChatColor = chatColors,
		arChatHistory = history,
		arSaveChannel = self.arSaveChannel,
		arFontOption = self.arFontOption,
		arKeyBindings = self.arKeyBindings,
		arChannelNames = self.arChannelNames,
		strSaveVersion = self.strSaveVersion
	}
	
	return tSaved
end

function ChatAdvanced:OnRestore(eType, tSavedData)
	
	if not self.arChatHistory then self.arChatHistory = {} end
	if not self.arSaveChannel then 
		self.arSaveChannel =
		{
			[ChatSystemLib.ChatChannel_Command] 		= 0,
			[ChatSystemLib.ChatChannel_System] 			= 0,
			[ChatSystemLib.ChatChannel_Debug] 			= 0,
			[ChatSystemLib.ChatChannel_Say] 			= 1,
			[ChatSystemLib.ChatChannel_Yell] 			= 1,
			[ChatSystemLib.ChatChannel_Whisper] 		= 1,
			[ChatSystemLib.ChatChannel_Party] 			= 1,
			[ChatSystemLib.ChatChannel_AnimatedEmote] 	= 1,
			[ChatSystemLib.ChatChannel_Zone] 			= 1,
			[ChatSystemLib.ChatChannel_ZonePvP] 		= 1,
			[ChatSystemLib.ChatChannel_Trade] 			= 1,
			[ChatSystemLib.ChatChannel_Guild] 			= 1,
			[ChatSystemLib.ChatChannel_GuildOfficer] 	= 1,
			[ChatSystemLib.ChatChannel_Society] 		= 1,
			[ChatSystemLib.ChatChannel_Custom] 			= 1,
			[ChatSystemLib.ChatChannel_NPCSay] 			= 0,
			[ChatSystemLib.ChatChannel_NPCYell] 		= 0,
			[ChatSystemLib.ChatChannel_NPCWhisper] 		= 0,
			[ChatSystemLib.ChatChannel_Datachron] 		= 0,
			[ChatSystemLib.ChatChannel_Combat] 			= 0,
			[ChatSystemLib.ChatChannel_Realm] 			= 0,
			[ChatSystemLib.ChatChannel_Loot] 			= 0,
			[ChatSystemLib.ChatChannel_Emote] 			= 1,
			[ChatSystemLib.ChatChannel_PlayerPath] 		= 0,
			[ChatSystemLib.ChatChannel_Instance] 		= 1,
			[ChatSystemLib.ChatChannel_WarParty] 		= 1,
			[ChatSystemLib.ChatChannel_WarPartyOfficer] = 1,
			[ChatSystemLib.ChatChannel_Advice] 			= 0,
			[ChatSystemLib.ChatChannel_AccountWhisper]	= 1,
		}
	end
	
	if not self.arFontOption then self.arFontOption = {} end
	if not self.arKeyBindings then 
		self.arKeyBindings = {
			{nCode=26, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.Keyboard},
			{nCode=309, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.Keyboard}
		} 
	end
	if not self.tOpenSlashCommand then
		self.tOpenSlashCommand = { 
			idAction=117, -- GameLib.CodeEnumInputAction.Unused9
			nDisplayIndex=4,
			idCategory=11, 
			strAction="OpenSlashCommand",
			strActionLocalized="Open Slash Command", -- Sorry to anyone with a non-English client 
			arInputs={
				{nCode=26, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.Keyboard},
				{nCode=309, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.Keyboard},
				{nCode=0, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.None},
				{nCode=0, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.None},
				{nCode=0, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.None},
				{nCode=0, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.None},
				{nCode=0, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.None},
				{nCode=0, eModifier=0, eDevice=GameLib.CodeEnumInputDevice.None} 
			}
		}

	end
	
	if not self.arChatColor then self.arChatColor = {} end
	if not self.arChannelNames then self.arChannelNames = {} end
			
	if tSavedData and tSavedData.strSaveVersion and tSavedData.strSaveVersion == self.strSaveVersion then
		if tSavedData.arChannelTypeToColor then
			for k,v in pairs(tSavedData.arChannelTypeToColor) do
				self.arChannelTypeToColor[k] = v
			end
		end
			
		if tSavedData.arChatColor then	
			for k,v in pairs(tSavedData.arChatColor) do
				self.arChatColor[k] = ApolloColor.new(v)	
			end
		end
			
		if tSavedData.arChatHistory then
			self.arChatHistory = tSavedData.arChatHistory
			self.bRestored = false
		end
		
		if tSavedData.arSaveChannel then
			self.arSaveChannel = tSavedData.arSaveChannel
		end
		
		if tSavedData.arFontOption then
			self.arFontOption = tSavedData.arFontOption
		end
		
		if tSavedData.arKeyBindings then
			self.arKeyBindings = tSavedData.arKeyBindings
		end
		
		if tSavedData.arChannelNames then
			self.arChannelNames = tSavedData.arChannelNames
		end
	end
	
	for i=1,#self.arKeyBindings do
		self.tOpenSlashCommand.arInputs[i] = self.arKeyBindings[i]
	end
	
	local Keybind = Apollo.GetAddon("Keybinding")
	Keybind.ChatAdvanced = self
	Keybind.OriginalPopulateKeybindList = Keybind.PopulateKeybindList
	Keybind.OriginalSave = Keybind.Save
		
	function Keybind:Save()   
		for idx = 1,#self.arKeyBindings do
			if self.arKeyBindings[idx].idAction == self.ChatAdvanced.tOpenSlashCommand.idAction then
				for i=1,2 do
					if self.arKeyBindings[idx].arInputs[i] then
						local keyname = GameLib.GetInputKeyNameText(self.arKeyBindings[idx].arInputs[i]):lower()
						if Key[keyname] then
							self.ChatAdvanced.arKeyBindings[i] = self.arKeyBindings[idx].arInputs[i]
						else
							self.ChatAdvanced.arKeyBindings[i] = nil
						end
					end
				end
				break
			end
		end

			
		self:OriginalSave()
	end
					
	function Keybind:PopulateKeybindList()
		table.insert(self.arKeyBindings, self.ChatAdvanced.tOpenSlashCommand)
		self:OriginalPopulateKeybindList()
	end
		
	local ChatLog = ChatAdvanced:GetChatLog()
	if ChatLog then
		if self.ChatLog.arChatColor then
			for k,v in pairs(self.arChatColor) do
				ChatLog.arChatColor[k] = v	
			end
			
			for k,v in pairs(ChatLog.arChatColor) do
				if(self.arSaveChannel[k] == nil) then
					self.arSaveChannel[k] = 1
				end
			end
		end
		

		ChatLog.OriginalHelperGenerateChatMessage = ChatLog.HelperGenerateChatMessage
		ChatLog.OriginalAddChannelTypeToList = ChatLog.AddChannelTypeToList
		ChatLog.OriginalOnChatMessage = ChatLog.OnChatMessage
		ChatLog.OriginalOnInputGainedFocus = ChatLog.OnInputGainedFocus
		ChatLog.OriginalOnInputChanged = ChatLog.OnInputChanged
		
		function ChatLog:OnInputGainedFocus(wndHandler, wndControl)
			if wndHandler == wndControl and wndHandler:GetParent() and wndHandler:GetParent():FindChild("InputType") then
				local wndForm = wndHandler:GetParent()
				local wndInputType = wndForm:FindChild("InputType")
				local wndInput = wndForm:FindChild("Input")
				local tChatData = wndForm:GetData()
				local channelInput = tChatData.channelCurrent
				local eChannelType = channelInput:GetType()
				local crText = self.arChatColor[channelInput:GetName()..channelInput:GetType()] or self.arChatColor[channelInput:GetType()] or ApolloColor.new("white")
				local strFontOption = self.strFontOption
				if self.ChatAdvanced.arFontOption and self.ChatAdvanced.arFontOption[eChannelName] then
					strFontOption = self.ChatAdvanced.arFontOption[eChannelName]
				elseif self.ChatAdvanced.arFontOption and self.ChatAdvanced.arFontOption[eChannelType] then
					strFontOption = self.ChatAdvanced.arFontOption[eChannelType]
				end 
				wndInputType:SetFont(strFontOption)
				wndInputType:SetTextColor(crText)
				wndInput:SetFont(strFontOption)
				wndInput:SetTextColor(crText)
				
				local offsets = {wndInput:GetAnchorOffsets()}
				wndInput:SetAnchorOffsets(offsets[1], -30, offsets[3], 0)

				wndHandler:GetParent():FindChild("InputType"):Show(string.len(wndHandler:GetParent():FindChild("Input"):GetText()) == 0)
				
				if self.tLastWhisperer and (eChannelType == ChatSystemLib.ChatChannel_Whisper or eChannelType == ChatSystemLib.ChatChannel_AccountWhisper) then
					local str = self.tLastWhisperer.strCharacterName
					if self.tLastWhisperer.strRealmName then
						str = str .. "@" .. self.tLastWhisperer.strRealmName
					end
					
					if eChannelType == ChatSystemLib.ChatChannel_AccountWhisper then
						str = "/aw " .. str .. " "
					else
						str = "/w " .. str .. " "
					end
							
					wndHandler:GetParent():FindChild("Input"):SetText(str)
					wndHandler:GetParent():FindChild("Input"):SetSel(str:len(), -1)
					wndHandler:GetParent():FindChild("InputType"):SetText("")
				end
			end
		end
		
		function ChatLog:OnInputChanged(wndHandler, wndControl, strText)
			local wndForm = wndControl:GetParent()
		
			if wndControl:GetName() ~= "Input" then
				return
			end
		
			for idx, wndChat in pairs(self.tChatWindows) do
				wndChat:FindChild("Input"):SetData(false)
			end
			wndControl:SetData(true)
		
			local wndForm = wndControl:GetParent()
			local wndInputType = wndForm:FindChild("InputType")
			local wndInput = wndForm:FindChild("Input")
			wndInputType:Show(string.len(strText) == 0) -- Hide background say once a message has been typed
		
			if strText == Apollo.GetString("ChatLog_Reply") and self.tLastWhisperer and self.tLastWhisperer.strCharacterName ~= "" then
				local strName = self.tLastWhisperer.strCharacterName
				local channel = self.channelWhisper
				if self.tLastWhisperer.eChannelType == ChatSystemLib.ChatChannel_AccountWhisper then
					channel = self.channelAccountWhisper
		
					self.tAccountWhisperContex =
					{
						["strDisplayName"]		= self.tLastWhisperer.strDisplayName,
						["strCharacterName"]	= self.tLastWhisperer.strCharacterName,
						["strRealmName"]		= self.tLastWhisperer.strRealmName,
					}
					strName = self.tLastWhisperer.strDisplayName
				end
		
				local strWhisper = String_GetWeaselString(Apollo.GetString("ChatLog_MessageToPlayer"), channel:GetAbbreviation(), strName)
		
				wndInputType:SetText(channel:GetCommand())
				wndInputType:SetTextColor(self.arChatColor[self.tLastWhisperer.eChannelType])
				wndInput:SetTextColor(self.arChatColor[self.tLastWhisperer.eChannelType])
				wndInput:SetText(strWhisper)
				wndInput:SetFocus()
				wndInput:SetSel(strWhisper:len(), -1)
				return
			end
		
			local tChatData = wndForm:GetData()
			if not strText then return end
			local tInput = ChatSystemLib.SplitInput(strText)
			local channelInput = tInput.channelCommand or tChatData.channelCurrent
			local eChannelType = channelInput:GetType()
			local crText = self.arChatColor[channelInput:GetName()..channelInput:GetType()] or self.arChatColor[channelInput:GetType()] or ApolloColor.new("white")

			local strFontOption = self.strFontOption
			if self.ChatAdvanced.arFontOption and self.ChatAdvanced.arFontOption[channelInput:GetName()..channelInput:GetType()] then
				strFontOption = self.ChatAdvanced.arFontOption[channelInput:GetName()..channelInput:GetType()]
			elseif self.ChatAdvanced.arFontOption and self.ChatAdvanced.arFontOption[channelInput:GetType()] then
				strFontOption = self.ChatAdvanced.arFontOption[channelInput:GetType()]
			end 
			wndInputType:SetFont(strFontOption)
			wndInputType:SetTextColor(crText)
			wndInput:SetFont(strFontOption)
			wndInput:SetTextColor(crText)
		
			if channelInput:GetType() == ChatSystemLib.ChatChannel_Command then -- command or emote
				if tInput.bValidCommand then
					wndInputType:SetText(String_GetWeaselString(Apollo.GetString("CRB_CurlyBrackets"), "", tInput.strCommand))
					wndInput:SetTextColor(kcrValidColor)
					wndInputType:SetTextColor(kcrValidColor)
				else
					wndInputType:SetText("X")
					wndInputType:SetTextColor(kcrInvalidColor)
				end
			else -- chatting in a channel; check for visibility
				--if tChatData.tViewedChannels[ channel:GetType() ] ~= nil then -- channel is viewed
				if self.tAllViewedChannels[ channelInput:GetType() ] ~= nil then -- channel is viewed
					wndInputType:SetText(channelInput:GetCommand())
				else -- channel is hidden
					wndInputType:SetText(String_GetWeaselString(Apollo.GetString("ChatLog_Invalid"), channelInput:GetCommand()))
					wndInputType:SetTextColor(kcrInvalidColor)
				end
			end
		end
							
		function ChatLog:OnChatMessage(channelCurrent, tMessage)
			-- tMessage has bAutoResponse, bGM, bSelf, strSender, strRealmName, nPresenceState, arMessageSegments, unitSource, bShowChatBubble, bCrossFaction, nReportId
		
			-- arMessageSegments is an array of tables.  Each table representsa part of the message + the formatting for that segment.
			-- This allows us to signal font (alien text for example) changes mid stream.
			-- local example = arMessageSegments[1]
			-- example.strText is the text
			-- example.bAlien == true if alien font set
			-- example.bRolePlay == true if this is rolePlay Text.  RolePlay text should only show up for people in roleplay mode, and non roleplay text should only show up for people outside it.
		
			-- to use: 	{#}toggles alien on {*}toggles rp on. Alien is still on {!}resets all format codes.
		
		
			-- There will be a lot of chat messages, particularly for combat log.  If you make your own chat log module, you will want to batch
			-- up several at a time and only process lines you expect to see.
			
			
		
			local tQueuedMessage = {}
			tQueuedMessage.tMessage = tMessage
			tQueuedMessage.eChannelType = channelCurrent:GetType()
			tQueuedMessage.strChannelName = channelCurrent:GetName()
			tQueuedMessage.strChannelCommand = channelCurrent:GetCommand()
		
			-- handle unit bubble if needed.
			if tQueuedMessage.tMessage.unitSource and tQueuedMessage.tMessage.bShowChatBubble then
				self:HelperGenerateChatMessage(tQueuedMessage)
				if tQueuedMessage.xmlBubble then
					tMessage.unitSource:AddTextBubble(tQueuedMessage.xmlBubble)
				end
			end
			
			if(self.ChatAdvanced.arSaveChannel[channelCurrent:GetName()..channelCurrent:GetType()] or self.ChatAdvanced.arSaveChannel[channelCurrent:GetType()]) then
				local bSkip = false
				local level = self.ChatAdvanced.arSaveChannel[channelCurrent:GetName()..channelCurrent:GetType()] or self.ChatAdvanced.arSaveChannel[channelCurrent:GetType()]
				
				if level < 2 then
					if tQueuedMessage.tMessage.strSender == "" then
						for k,v in pairs(tQueuedMessage.tMessage.arMessageSegments) do
							if string.match(v.strText, "has gone offline") or string.match(v.strText, "has come online") then
								bSkip = true
							end
						end
					end
				end

				if not bSkip then
					table.insert(self.ChatAdvanced.arChatHistory, tQueuedMessage)
				end
			end
					
			-- queue message on windows.
			for key, wndChat in pairs(self.tChatWindows) do
				if wndChat:GetData().tViewedChannels[tQueuedMessage.eChannelType] then -- check flags for filtering
					self.bQueuedMessages = true
					wndChat:GetData().tMessageQueue:Push(tQueuedMessage)
				end
			end
		end
				
		function ChatLog:AddChannelTypeToList(tData, wndList, channel)
			local wndChannelItem = Apollo.LoadForm(self.xmlDoc, "ChatType", wndList, self)
			
			wndChannelItem:FindChild("TypeName"):SetText(channel:GetName())
			wndChannelItem:FindChild("TypeName"):SetAnchorOffsets(60,0,-60,0)
			wndChannelItem:SetData(channel:GetType())
			wndChannelItem:FindChild("ViewCheck"):SetCheck(tData.tViewedChannels[channel:GetType()] or false)
			wndChannelItem:FindChild("HoldCheck"):SetCheck(tData.tHeldChannels[channel:GetType()] or false)	
			
			-- Load the Pixie from the ChatAdvanced forms library
			color = self.arChatColor[channel:GetName()..channel:GetType()]	or self.arChatColor[channel:GetType()]
			local wndChannelColor = Apollo.LoadForm(self.ChatAdvanced.xmlDoc, "ChannelColor", wndChannelItem, self.ChatAdvanced)
			tPixieInfo = wndChannelColor:GetPixieInfo(1)
			tPixieInfo.cr = color
			wndChannelColor:UpdatePixie(1, tPixieInfo)	
			wndChannelColor:SetData({ oChannel = channel, nPixieId = 1 })
		end
					
		-- And now for the main event, replace the HelperGenerateChatMessage
		-- This is necessary since the original uses the karChannelTypeToColor
		-- table, which is otherwise inaccessible outside of ChatLog's scope
		function ChatLog.HelperGenerateChatMessage(tSelf, tQueuedMessage)
			if tQueuedMessage.xml then
				return
			end
	
			local eChannelType = tQueuedMessage.eChannelType
			local strChannelName = tQueuedMessage.strChannelName
			local eChannelName = strChannelName..eChannelType
			local tMessage = tQueuedMessage.tMessage
	
			-- Different handling for combat log
			if eChannelType == ChatSystemLib.ChatChannel_Combat then
				-- no formats in combat, roll it all up into one.
				local strMessage = ""
				for idx, tSegment in ipairs(tMessage.arMessageSegments) do
					strMessage = strMessage .. tSegment.strText
				end
				tQueuedMessage.strMessage = strMessage
				return
			end
	
			local xml = XmlDoc.new()
			local tm = GameLib.GetLocalTime()
			local crText = tSelf.arChatColor[eChannelName] or tSelf.arChatColor[eChannelType] or ApolloColor.new("white") -- Make sure that we use ChatAdvanced's color tables
			local strChannelColor = "white"
			if tSelf.ChatAdvanced.arChannelTypeToColor[eChannelName] then
				strChannelColor = tSelf.ChatAdvanced.arChannelTypeToColor[eChannelName].Channel
			elseif tSelf.ChatAdvanced.arChannelTypeToColor[eChannelType] then
				strChannelColor = tSelf.ChatAdvanced.arChannelTypeToColor[eChannelType].Channel
			end
			
			local crChannel = ApolloColor.new(strChannelColor)
			local strFontOption = tSelf.strFontOption
			if tSelf.ChatAdvanced.arFontOption and tSelf.ChatAdvanced.arFontOption[eChannelName] then
				strFontOption = tSelf.ChatAdvanced.arFontOption[eChannelName]
			elseif tSelf.ChatAdvanced.arFontOption and tSelf.ChatAdvanced.arFontOption[eChannelType] then
				strFontOption = tSelf.ChatAdvanced.arFontOption[eChannelType]

			end 
			local crPlayerName = ApolloColor.new("ChatPlayerName")
						
			local strTime = "" if tSelf.bShowTimestamp then strTime = string.format("%d:%02d ", tm.nHour, tm.nMinute) end
			local strWhisperName = tMessage.strSender
			if tMessage.strRealmName:len() > 0 then
				strWhisperName = strWhisperName .. "@" .. tMessage.strRealmName
			end
	
			local strPresenceState = ""
			if tMessage.bAutoResponse then
				strPresenceState = '('..Apollo.GetString("AutoResponse_Prefix")..')'
			end
	
			if tMessage.nPresenceState == FriendshipLib.AccountPresenceState_Away then
				strPresenceState = '<'..Apollo.GetString("Command_Friendship_AwayFromKeyboard")..'>'
			elseif tMessage.nPresenceState == FriendshipLib.AccountPresenceState_Busy then
				strPresenceState = '<'..Apollo.GetString("Command_Friendship_DoNotDisturb")..'>'
			end
	
			if eChannelType == ChatSystemLib.ChatChannel_Whisper then
				if not tMessage.bSelf then
					tSelf.tLastWhisperer = { strCharacterName = strWhisperName, eChannelType = ChatSystemLib.ChatChannel_Whisper }
				end
				Sound.Play(Sound.PlayUISocialWhisper)
			elseif eChannelType == ChatSystemLib.ChatChannel_AccountWhisper then
	
				local tPreviousWhisperer = tSelf.tLastWhisperer
	
				tSelf.tLastWhisperer =
				{
					strCharacterName = tMessage.strSender,
					strRealmName = nil,
					strDisplayName = nil,
					eChannelType = ChatSystemLib.ChatChannel_AccountWhisper
				}
	
				local tAccountFriends = FriendshipLib.GetAccountList()
				for idx, tAccountFriend in pairs(tAccountFriends) do
					if tAccountFriend.arCharacters ~= nil then
						for idx, tCharacter in pairs(tAccountFriend.arCharacters) do
							if tCharacter.strCharacterName == tMessage.strSender and (tMessage.strRealmName:len() == 0 or tCharacter.strRealm == tMessage.strRealmName) then
								if not tMessage.bSelf or (tPreviousWhisperer and tPreviousWhisperer.strCharacterName == tMessage.strSender) then
									tSelf.tLastWhisperer.strDisplayName = tAccountFriend.strCharacterName
									tSelf.tLastWhisperer.strRealmName = tCharacter.strRealm
								end
								strWhisperName = tAccountFriend.strCharacterName
								if tMessage.strRealmName:len() > 0 then
									strWhisperName = strWhisperName .. "@" .. tMessage.strRealmName
								end
							end
						end
					end
				end
				Sound.Play(Sound.PlayUISocialWhisper)
			end
	
			if eChannelType == ChatSystemLib.ChatChannel_AnimatedEmote then
				xml:AddLine(strTime, crChannel, strFontOption, "Left")
	
			elseif eChannelType == ChatSystemLib.ChatChannel_Emote then
				xml:AddLine(strTime, crChannel, strFontOption, "Left")
				if strWhisperName:len() > 0 then
					if tMessage.bGM then
						xml:AppendImage(tSelf.strGMIcon, 16, 16)
					end
					xml:AppendText(strWhisperName, crPlayerName, strFontOption, {CharacterName=strWhisperName, nReportId=tMessage.nReportId}, "Source")
				end
				xml:AppendText(" ")
			else
				local strChannelName = tSelf.ChatAdvanced.arChannelNames[eChannelName] or tQueuedMessage.strChannelName
			
				if eChannelType == ChatSystemLib.ChatChannel_Society then
					strChannel = String_GetWeaselString(Apollo.GetString("ChatLog_GuildCommand"), strChannelName, tQueuedMessage.strChannelCommand).." "
				else
					strChannel = String_GetWeaselString(Apollo.GetString("CRB_Brackets_Space"), strChannelName)
				end
		
				if tSelf.bShowChannel ~= true then
					strChannel = ""
				end
	
				xml:AddLine(strTime .. strChannel, crChannel, strFontOption, "Left")
				if strWhisperName:len() > 0 then
	
					local strWhisperNamePrefix = ""
					if eChannelType == ChatSystemLib.ChatChannel_Whisper or eChannelType == ChatSystemLib.ChatChannel_AccountWhisper then
						if tMessage.bSelf then
							strWhisperNamePrefix = Apollo.GetString("ChatLog_To")
						else
							strWhisperNamePrefix = Apollo.GetString("ChatLog_From")
						end
					end
	
					xml:AppendText( strWhisperNamePrefix, crText, strFontOption)
	
					if tMessage.bGM then
						xml:AppendImage(tSelf.strGMIcon, 16, 16)
					end
	
					xml:AppendText( strWhisperName, crPlayerName, strFontOption, {CharacterName=strWhisperName, nReportId=tMessage.nReportId}, "Source")
				end
				xml:AppendText( strPresenceState .. ": ", crChannel, strFontOption, "Left")
			end
	
			local xmlBubble = nil
			if tMessage.bShowChatBubble then
				xmlBubble = XmlDoc.new()
				xmlBubble:AddLine("", crChannel, strFontOption, "Center")
			end
	
			local bHasVisibleText = false
			for idx, tSegment in ipairs( tMessage.arMessageSegments ) do
				local strText = tSegment.strText
				local bAlien = tSegment.bAlien or tMessage.bCrossFaction
				local bShow = false
	
				if tSelf.eRoleplayOption == 3 then
					bShow = not tSegment.bRolePlay
				elseif tSelf.eRoleplayOption == 2 then
					bShow = tSegment.bRolePlay
				else
					bShow = true;
				end
	
				if bShow then
					local crChatText = crText;
					local crBubbleText = tSelf.strColorChatRegular
					local strChatFont = strFontOption
					local strBubbleFont = tSelf.strBubbleFont
					local tLink = {}
	
	
					if tSegment.uItem ~= nil then
						strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), tSegment.uItem:GetName())
						crChatText = tSelf.arEvalColors[tSegment.uItem:GetItemQuality()]
						crBubbleText = ApolloColor.new("white")
	
						tLink.strText = strText
						tLink.uItem = tSegment.uItem
	
					elseif tSegment.uQuest ~= nil then
						strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), tSegment.uQuest:GetTitle())
						crChatText = ApolloColor.new("green")
						crBubbleText = ApolloColor.new("green")
	
						tLink.strText = strText
						tLink.uQuest = tSegment.uQuest
	
					elseif tSegment.uArchiveArticle ~= nil then
						strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), tSegment.uArchiveArticle:GetTitle())
						crChatText = ApolloColor.new("ffb7a767")
						crBubbleText = ApolloColor.new("ffb7a767")
	
						tLink.strText = strText
						tLink.uArchiveArticle = tSegment.uArchiveArticle
	
					else
						if tSegment.bRolePlay then
							crBubbleText = tSelf.strColorChatRoleplay
							strChatFont = tSelf.strRPFontOption
							strBubbleFont = tSelf.strDialogFontRP
						end
	
						if bAlien or tSegment.bProfanity then
							strChatFont = tSelf.strAlienFontOption
							strBubbleFont = tSelf.strAlienFontOption
						end
					end
	
					if next(tLink) == nil then
						xml:AppendText(strText, crChatText, strChatFont)
					else
						local strLinkIndex = tostring( tSelf:HelperSaveLink(tLink) )
						xml:AppendText(strText, crChatText, strChatFont, {strIndex=strLinkIndex} , "Link")
					end
	
					if xmlBubble then
						xmlBubble:AppendText(strText, crBubbleText, strBubbleFont)
					end
	
					bHasVisibleText = bHasVisibleText or tSelf:HelperCheckForEmptyString(strText)
				end
			end
	
			tQueuedMessage.bHasVisibleText = bHasVisibleText
			tQueuedMessage.xml = xml
			tQueuedMessage.xmlBubble = xmlBubble
		end		
	end	

end

-----------------------------------------------------------------------------------------------
-- ChatAdvanced OnDocLoaded
-----------------------------------------------------------------------------------------------
function ChatAdvanced:OnDocLoaded()

end

---------------------------------------------------------------------------------------------------
-- ColorPickerForm Functions
---------------------------------------------------------------------------------------------------

function ChatAdvanced:OnSaveChatLogCheck( wndHandler, wndControl, eMouseButton )
end

function ChatAdvanced:OnSaveChatLogUncheck( wndHandler, wndControl, eMouseButton )
end


-----------------------------------------------------------------------------------------------
-- ChatAdvanced Instance
-----------------------------------------------------------------------------------------------
local ChatAdvancedInst = ChatAdvanced:new()
ChatAdvancedInst:Init()


-----------------------------------------------------------------------------------------------
-- ColorPicker Plugin
-----------------------------------------------------------------------------------------------
-- Authors: Jaslm and Tomber, all rights reserved
-- Modified by Kertsz to better support integration

ColorPickerAdvanced = {}
local lastTime, updateTime = 0, 0 -- used to throttle updates triggered by mouse moves
local updateThrottle = 0.05 -- time in seconds between updates
local mouseInRing, mouseInGradient = false, false -- keep track of where mouse was clicked
local focusWidth, focusHeight, ringWidth, ringHeight, gradientWidth, gradientHeight, graderWidth, graderHeight
local selectWidth, selectHeight, rsquareMin, rsquareMax, radius
local ringEdge = 2 -- number of pixels ring is inset in the form
local ringThick = 25 -- thickness of the ring in pixels
local hue, saturation, brightness, alpha = 0, 1, 0.5, 1 -- current color value in HSV
local showAlpha = true -- if false hide the alpha-related controls and act like alpha is 1.0
local oldColor = CColor.new(1, 0, 0, 1) -- default first time open with slash command
local newColor = CColor.new(1, 1, 1, 1) -- set to the color that is being updated, overwriting this init value
local oldFont = nil
local ccColor = {} -- buffer for copy and paste function
local ccValid = false -- don't display the copy buffer until it is valid
local c_white = CColor.new(1, 1, 1, 1)
local c_select = CColor.new(1, 0, 0, 1)
local callback = nil
local parameter = nil
local target = nil -- Added a callback target

-- Convert r, g, b input values into h, s, l return values
local function Convert_RGB_To_HSV(r, g, b)
	local mincolor, maxcolor = math.min(r, g, b), math.max(r, g, b)
	local ch, cs, cv = 0, 0, maxcolor
	if maxcolor > 0 then -- technically ch is undefined if cs is zero
		local delta = maxcolor - mincolor
		cs = delta / maxcolor
		if delta > 0 then -- don't allow divide by zero
			if r == maxcolor then
				ch = (g - b) / delta -- between yellow and magenta
			elseif g == maxcolor then
				ch = 2 + ((b - r) / delta) -- between cyan and yellow
			else
				ch = 4 + ((r - g) / delta) -- between magenta and cyan
			end
		end
		if ch < 0 then ch = ch + 6 end -- correct for negative values
		ch = ch / 6 -- and finally adjust range 0 to 1.0
	end
	return ch, cs, cv
end

-- Convert h, s, l input values into r, g, b return values
-- All values are in the range 0 to 1.0
local function Convert_HSV_To_RGB(ch, cs, cv)
	local r, g, b = cv, cv, cv
	if cs > 0 then -- if cs is zero then grey is returned
		local h = ch * 6; local sextant = math.floor(h) -- figure out which sextant of the color wheel
		local fract = h - sextant -- fractional offset into the sextant
		local p, q, t = cv * (1 - cs), cv * (1 - (cs * fract)), cv * (1 - (cs * (1 - fract)))
		if sextant == 0 then
			r, g, b = cv, t, p
		elseif sextant == 1 then
			r, g, b = q, cv, p
		elseif sextant == 2 then
			r, g, b = p, cv, t
		elseif sextant == 3 then
			r, g, b = p, q, cv
		elseif sextant == 4 then
			r, g, b = t, p, cv
		else
			r, g, b = cv, p, q
		end
	end
	return r, g, b
end

-- Convert r, g, b values (in range 0 to 1.0) into a hex-encoded string using format "rrggbb"
local function Convert_RGB_To_String(r, g, b)
	return string.format("%02x%02x%02x", math.floor(r * 255 + 0.5), math.floor(g * 255 + 0.5), math.floor(b * 255 + 0.5))
end

-- Convert a hex-encoded string into r, g, b values, string format is assumed to be "rrggbb"
local function Convert_String_To_RGB(hex)
	local r, g, b = 0, 0, 0 -- invalid strings will result in these values being returned
	local n = tonumber(hex, 16)
	if n then r = math.floor(n / 65536); g = math.floor(n / 256) % 256; b = n % 256 end
	return r / 255, g / 255, b / 255
end

-- Convert a color value in range 0 to 1.0 to string for range 0 to 255
local function CT(x) return tostring(math.floor(x * 255 + 0.5)) end

-- Addon initialization and registration
function ColorPickerAdvanced:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    return o
end

-- Update values in controls from current HSV value, assume already initialized cached form handlers
local function UpdateColorControls()
	local r, g, b = Convert_HSV_To_RGB(hue, saturation, brightness)
	local cp = ColorPickerAdvanced
	cp.select:SetBGColor(mouseInRing and c_select or c_white) -- display focus indicator when changing ring
	cp.grader:SetBGColor(mouseInGradient and c_select or c_white) -- display focus indicator when changing gradient
	
	cp.oldTop:SetBGColor(CColor.new(oldColor.r, oldColor.g, oldColor.b, oldColor.a))
	cp.oldBottom:SetBGColor(CColor.new(oldColor.r, oldColor.g, oldColor.b, oldColor.a))
	cp.newColor:SetBGColor(CColor.new(r, g, b, alpha)) -- use color passed in thru API
	
	newColor.r = r; newColor.g = g; newColor.b = b; newColor.a = alpha -- update the current color value
	
	cp.red:SetText(CT(r)); cp.green:SetText(CT(g)); cp.blue:SetText(CT(b))
	local a = math.floor(alpha * 100 + 0.5) -- round alpha to nearest integer 0 to 100
	cp.alpha:SetText(string.format("%0.0f", a)); cp.slider:SetValue(a)
	cp.hex:SetText(Convert_RGB_To_String(r, g, b))
	
	r, g, b = Convert_HSV_To_RGB(hue, 1, 1)
	cp.olay:SetBGColor(CColor.new(r, g, b, 1))
	local ch = hue -- use this to adjust color selection point
	ch = hue - 0.25; if ch < 0 then ch = ch + 1 end -- rotate to correct orientation
	local theta = ch * 2 * math.pi -- convert to radians
	local offL, offT, offR, offB = cp.ring:GetAnchorOffsets()
	local cx = radius * math.cos(theta)
	local cy = radius * math.sin(theta)
	local dL = offL + ringWidth + cx
	local dT = offT + ringHeight + cy
	offL = dL - selectWidth + ringEdge; offT = dT - selectHeight + ringEdge
	offR = dL + selectWidth + ringEdge; offB = dT + selectHeight + ringEdge
	cp.select:SetAnchorOffsets(offL, offT, offR, offB)

	local rx = ((saturation * 2) - 1) * gradientWidth -- use saturation and brightness to adjust gradient selection point
	local ry = (((brightness - 1) * -2) - 1) * gradientHeight
	dL = gradientWidth + rx; dT = gradientHeight + ry
	offL = dL - graderWidth; offT = dT - graderHeight
	offR = dL + graderWidth; offB = dT + graderHeight
	cp.grader:SetAnchorOffsets(offL, offT, offR, offB)
	
	cp.hint:Show(Apollo.IsControlKeyDown()) -- display hint when the control key is held down
	cp.copy:Show(ccValid); cp.paste:Show(ccValid) -- only display the copy buffer and paste button after a copy has been made
	if ccValid then
		r, g, b = Convert_HSV_To_RGB(ccColor.hue, ccColor.saturation, ccColor.brightness)
		cp.copy:SetBGColor(CColor.new(r, g, b, 1))
	end
	
	cp.alphabox:Show(showAlpha); cp.slider:Show(showAlpha); cp.sframe:Show(showAlpha); cp.back:Show(showAlpha)
	if type(callback) == "function" then callback(target, parameter, {color = newColor, font = oldFont}, false) end -- Modified to also pass the new color and whether or not this was triggered by clicking Ok
end

-- Clear the focus from all the text boxes
local function ClearFocus()
	local cp = ColorPickerAdvanced
	cp.red:ClearFocus(); cp.green:ClearFocus(); cp.blue:ClearFocus(); cp.hex:ClearFocus(); cp.alpha:ClearFocus()
end

-- At load time, plug in our custom images
function ColorPickerAdvanced:OnLoad()
	Apollo.RegisterSlashCommand("ColorPickerAdvanced", "OnColorPickerAdvancedOn", self)
	Apollo.RegisterSlashCommand("coloradvanced", "OnColorPickerAdvancedOn", self)

	self.wndMain = Apollo.LoadForm("ChatAdvanced.xml", "ColorPickerForm", nil, self)
	self.wndMain:Show(false) -- hide the window by default
	Apollo.LoadSprites("ColorPickerAdvancedSprites.xml", "ColorPickerAdvancedSprites")
	callback = nil; parameter = nil

	local cp = ColorPickerAdvanced
	cp.window = self.wndMain
	cp.ring = self.wndMain:FindChild("ColorRing")
	cp.rframe = self.wndMain:FindChild("ColorRingFrame")
	cp.olay = self.wndMain:FindChild("ColorOverlay")
	cp.grad = self.wndMain:FindChild("ColorGradient")
	cp.gframe = self.wndMain:FindChild("ColorGradientFrame")
	cp.back = self.wndMain:FindChild("Checkerboard")
	cp.select = self.wndMain:FindChild("ColorSelector")
	cp.oldTop = self.wndMain:FindChild("OldColorTop")
	cp.oldBottom = self.wndMain:FindChild("OldColorBottom")
	cp.newColor = self.wndMain:FindChild("NewColor")
	cp.grader = cp.olay:FindChild("ColorSelector")
	cp.red = self.wndMain:FindChild("RedBox"):FindChild("Editor")
	cp.green = self.wndMain:FindChild("GreenBox"):FindChild("Editor")
	cp.blue = self.wndMain:FindChild("BlueBox"):FindChild("Editor")
	cp.hex = self.wndMain:FindChild("HexBox"):FindChild("Editor")
	cp.alphabox = self.wndMain:FindChild("AlphaBox")
	cp.alpha = cp.alphabox:FindChild("Editor")
	cp.slider = self.wndMain:FindChild("AlphaSlider")
	cp.sframe = self.wndMain:FindChild("AlphaBackground")
	cp.hint = self.wndMain:FindChild("Hint")
	cp.copy = self.wndMain:FindChild("CopyColor")
	cp.paste = self.wndMain:FindChild("PasteButton")
	assert(cp.ring and cp.rframe and cp.olay and cp.grad and cp.gframe and cp.back and cp.select and cp.grader and cp.slider and cp.sframe and
		cp.oldTop and cp.oldBottom and cp.newColor and cp.red and cp.green and cp.blue and cp.hex and cp.alpha and cp.alphabox and cp.hint and cp.copy and cp.paste,
		"ColorPickerAdvanced: form initialization error")

	cp.ring:SetSprite("ColorPickerAdvancedSprites:sprColorRing")
	cp.grad:SetSprite("ColorPickerAdvancedSprites:sprColorGradient")
	cp.olay:SetSprite("ColorPickerAdvancedSprites:sprColorOverlay")
	cp.olay:SetBGColor(CColor.new(1, 0, 0, 1))
	cp.back:SetSprite("ColorPickerAdvancedSprites:sprCheckerboard")

	focusWidth = cp.rframe:GetWidth() / 2; focusHeight = cp.rframe:GetHeight() / 2
	ringWidth = cp.ring:GetWidth() / 2; ringHeight = cp.ring:GetHeight() / 2
	local w = ringWidth - ringEdge; rsquareMax = (w * w); radius = w - (ringThick / 2) - ringEdge
	w = w - ringThick; rsquareMin = (w * w)
	gradientWidth = cp.olay:GetWidth() / 2; gradientHeight = cp.olay:GetHeight() / 2
	graderWidth = cp.grader:GetWidth() / 2; graderHeight = cp.grader:GetHeight() / 2
	selectWidth = cp.select:GetWidth() / 2; selectHeight = cp.select:GetHeight() / 2

	UpdateColorControls() -- set initial value of HSV into the controls
end

-- API to display the color picker dialog for a color object, optionally with alpha-related controls
-- Don't return until the dialog is closed, restore original color if cancelled
-- The function cb is called with the parameter pm whenever the color may have been changed (optional)
function ColorPickerAdvanced.AdjustCColor(color, hasAlpha, cb, pm, tg, data)
	local cp = ColorPickerAdvanced
	oldColor.r = color.r; oldColor.g = color.g; oldColor.b = color.b; oldColor.a = color.a
	newColor = color
	callback = cb; parameter = pm; target = tg; oldFont = data.font or "CRB_InterfaceMedium"
	cp.window:FindChild("FontSelector"):SetFont(oldFont)
	cp.window:FindChild("FontSelector"):SetData(oldFont)
	local flist = cp.window:FindChild("FontList"):FindChild(oldFont)
	local fname = flist:GetText()
	cp.window:FindChild("FontSelector"):SetText(fname)
	hue, saturation, brightness = Convert_RGB_To_HSV(color.r, color.g, color.b)
	if hasAlpha then alpha = color.a else alpha = 1 end
	showAlpha = hasAlpha
	UpdateColorControls() -- update current color value in controls
	cp.window:Show(true) -- show the window
	cp.window:ToFront() -- bring it to the front
end

-- Respond to registered slash commands
function ColorPickerAdvanced:OnColorPickerAdvancedOn()
	hue, saturation, brightness = Convert_RGB_To_HSV(oldColor.r, oldColor.g, oldColor.b)
	alpha = oldColor.a
	showAlpha = true
	newColor = CColor.new(1, 1, 1, 1)
	callback = nil; parameter = nil
	UpdateColorControls() -- update current color value in controls
	self.wndMain:Show(true) -- show the window
	self.wndMain:ToFront() -- bring it to the front
end

-- Process a click in the color or gradient selection areas
local function ColorPickerAdvancedMouseEvent(wndHandler, wndControl, event, x, y)
	local cp = ColorPickerAdvanced
	local now = GameLib.GetGameTime(); local elapsedTime = now - lastTime; lastTime = now
	local forced = false
	updateTime = updateTime + elapsedTime
	local rx, ry = x - focusWidth, y - focusHeight -- get mouse position as ring coordinates from center of focus
	local rsquare = (rx * rx) + (ry * ry)
	local inRing = (rsquare >= rsquareMin) and (rsquare <= rsquareMax) -- figure out if in the ring
	local inGradient = (math.abs(rx) <= gradientWidth) and (math.abs(ry) <= gradientHeight)
	if event == "down" then -- handle mouse clicks
		if inRing and not mouseInGradient then mouseInRing = true; forced = true end
		if inGradient and not mouseInRing then mouseInGradient = true; forced = true end
		ClearFocus()
	elseif (event == "up" or event == "exit") and (mouseInRing or mouseInGradient) then
		mouseInRing = false; mouseInGradient = false; forced = true
		UpdateColorControls() -- display current value in the controls
	end
	if forced or updateTime >= updateThrottle then
		if mouseInRing then
			local theta = math.atan2(ry, rx)
			if theta < 0 then theta = theta + (2 * math.pi) end
			hue = theta / (2 * math.pi) -- hue is angle in the range 0 to 1.0
			if Apollo.IsControlKeyDown() then hue = math.floor((hue * 24) + 0.5) / 24 end -- if control key then snap to nearest standard-ish color
			hue = hue + 0.25; if hue >= 1 then hue = hue - 1 end -- rotate to correct orientation
			UpdateColorControls() -- display current value in the controls
		elseif mouseInGradient then
			if rx > gradientWidth then rx = gradientWidth elseif rx < -gradientWidth then rx = -gradientWidth end
			if ry > gradientHeight then ry = gradientHeight elseif ry < -gradientHeight then ry = -gradientHeight end
			saturation = (1 + (rx / gradientWidth)) / 2
			brightness = 1 - ((1 + (ry / gradientHeight)) / 2)
			if Apollo.IsControlKeyDown() then -- if control key then snap to nearest 1/8
				saturation = math.floor((saturation * 8) + 0.5) / 8; brightness = math.floor((brightness * 8) + 0.5) / 8
			end
			UpdateColorControls() -- display current value in the controls
		end
		updateTime = 0; forced = false
	end
	cp.hint:Show(Apollo.IsControlKeyDown()) -- only display hint when the control key is being held down
end

-- Respond to OK button click
function ColorPickerAdvanced:OnOK()
	local r, g, b = Convert_HSV_To_RGB(hue, saturation, brightness)
	oldColor.r = r; oldColor.g = g; oldColor.b = b -- save current color in original color for next time
	if type(callback) == "function" then callback(target, parameter, { color = newColor, font = self.wndMain:FindChild("FontSelector"):GetData()} , true) end -- Same modifications as to UpdateColorControls
	callback = nil; parameter = nil; target = nil
	self.wndMain:Show(false) -- hide the window
end

function ColorPickerAdvanced:OnOKReload()
	local r, g, b = Convert_HSV_To_RGB(hue, saturation, brightness)
	oldColor.r = r; oldColor.g = g; oldColor.b = b -- save current color in original color for next time
	if type(callback) == "function" then callback(target, parameter, { color = newColor, font = self.wndMain:FindChild("FontSelector"):GetData()} , true) end -- Same modifications as to UpdateColorControls
	callback = nil; parameter = nil; target = nil
	self.wndMain:Show(false) -- hide the window
	RequestReloadUI()
end


-- Respond to Cancel button click
function ColorPickerAdvanced:OnCancel()
	newColor.r = oldColor.r; newColor.g = oldColor.g; newColor.b = oldColor.b; newColor.a = oldColor.a -- revert to original color
	if type(callback) == "function" then callback(target, parameter, { color = newColor, font = oldFont }, false) end -- Same modifications as to UpdateColorControls
	callback = nil; parameter = nil; target = nil
	self.wndMain:Show(false) -- hide the window
end

-- Respond to mouse button down in the color selection ring
function ColorPickerAdvanced:OnMouseDownColor(wndHandler, wndControl, key, x, y)
	if key == 0 then ColorPickerAdvancedMouseEvent(wndHandler, wndControl, "down", x, y) end
end

-- Respond to mouse move events in the color selection ring
function ColorPickerAdvanced:OnMouseMoveColor(wndHandler, wndControl, x, y)
	ColorPickerAdvancedMouseEvent(wndHandler, wndControl, "move", x, y)
end

-- Respond to mouse button up in the color selection ring
function ColorPickerAdvanced:OnMouseUpColor(wndHandler, wndControl, key, x, y)
	if key == 0 then ColorPickerAdvancedMouseEvent(wndHandler, wndControl, "up", x, y) end
end

-- Respond to mouse exit events in the color selection ring
function ColorPickerAdvanced:OnMouseExitColor(wndHandler, wndControl)
	ColorPickerAdvancedMouseEvent(wndHandler, wndControl, "exit", 0, 0)
end

-- Respond to text changes in the red, green or blue box, managing the selection and focus appropriately
-- Try to convert to a number, if valid and in range then update current setting
function ColorPickerAdvanced:OnColorChanged(wndHandler, wndControl, s)
	local cp = ColorPickerAdvanced
	local rflag, gflag, bflag = false, false, false
	local rs = cp.red:GetText(); if rs == "" then rs = "0"; rflag = true end
	local gs = cp.green:GetText(); if gs == "" then gs = "0"; gflag = true end
	local bs = cp.blue:GetText(); if bs == "" then bs = "0"; bflag = true end
	local r, g, b = tonumber(rs), tonumber(gs), tonumber(bs)
	if r and r >= 0 and r <= 255 and g and g >= 0 and g <= 255 and b and b >= 0 and b <= 255 then
		hue, saturation, brightness = Convert_RGB_To_HSV(r / 255, g / 255, b / 255)
	end
	UpdateColorControls() -- update current color values
	if rflag then cp.red:SetText("") else r = cp.red:GetText(); if r then cp.red:SetSel(r:len(), -1) end end
	if gflag then cp.green:SetText("") else g = cp.green:GetText(); if g then cp.green:SetSel(g:len(), -1) end end
	if bflag then cp.blue:SetText("") else b = cp.blue:GetText(); if b then cp.blue:SetSel(b:len(), -1) end end
end

-- Respond to text changes in the hex box, managing the selection and focus appropriately
-- Try to convert to a hex number, if valid length and format
function ColorPickerAdvanced:OnHexChanged(wndHandler, wndControl, s)
	local cp = ColorPickerAdvanced
	local ht = cp.hex:GetText()
	if not ht or ht:len() ~= 6 then return end
	local r, g, b = Convert_String_To_RGB(ht)
	hue, saturation, brightness = Convert_RGB_To_HSV(r, g, b)
	UpdateColorControls() -- update current color values
	cp.hex:SetSel(6, -1)
end

-- Respond to text changes in the alpha box, managing the selection and focus appropriately
-- Try to convert to a number in the range 0 to 100
function ColorPickerAdvanced:OnAlphaChanged(wndHandler, wndControl, s)
	local cp = ColorPickerAdvanced
	local aflag = false
	local at = cp.alpha:GetText(); if at == "" then at = "0"; aflag = true end
	local a = tonumber(at)
	if a then alpha = a / 100; if alpha < 0 then alpha = 0 elseif alpha > 1 then alpha = 1 end end
	UpdateColorControls() -- update current color values
	if aflag then cp.alpha:SetText("") else at = cp.alpha:GetText(); if at then cp.alpha:SetSel(at:len(), -1) end end
end

-- Respond to alpha slider changes
function ColorPickerAdvanced:OnAlphaChanging(wndHandler, wndControl, x)
	alpha = x / 100
	UpdateColorControls() -- update current color values
	ClearFocus()
end

-- Respond to copy and paste buttons
function ColorPickerAdvanced:OnCopyButton()
	ccColor.hue = hue; ccColor.saturation = saturation; ccColor.brightness = brightness -- save the current color in the copy buffer
	ccValid = true; UpdateColorControls(); ClearFocus()
end

function ColorPickerAdvanced:OnPasteButton()
	if ccValid then hue = ccColor.hue; saturation = ccColor.saturation; brightness = ccColor.brightness end -- set the current color to the copy buffer
	UpdateColorControls(); ClearFocus()
end

function ColorPickerAdvanced:ChangeFontClick( wndHandler, wndControl, eMouseButton )
	self.wndMain:FindChild("FontList"):Show(true)
end

function ColorPickerAdvanced:SetFont( wndHandler, wndControl, eMouseButton )
	self.wndMain:FindChild("FontSelector"):SetFont(wndControl:GetName())
	self.wndMain:FindChild("FontSelector"):SetData(wndControl:GetName())
	self.wndMain:FindChild("FontSelector"):SetText(wndControl:GetText())
	self.wndMain:FindChild("FontList"):Show(false)	
end


